/* 
 * Copyright (C) 2008 Miroslav Jezbera <jezz@hkfree.org>
 * This file is part of LQueues - simple bandwidth manager for linux.
 *
 * LQueues is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * LQueues is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with LQueues.  If not, see <http://www.gnu.org/licenses/agpl.html>.
 */

var speedHistoryLength = 10;
var nextCounter = 0;

/*
 * Queue object
 */
function Queue()
	{
	this.lastCounter = -1;
	this.lastUpdateTime = 0;
	this.bytes = 0;
	this.speeds = reallocateArray(0, speedHistoryLength);
	this.rate = 0;
	}

Queue.prototype.assignInfoElements = function ($speedInfo, $rateInfo)
	{
	this.$speedInfo = $speedInfo;
	this.$rateInfo = $rateInfo;
	};

Queue.prototype.update = function (counter, bytes, rate, time)
	{
	if (counter <= this.lastCounter)
		return;
	var speed = (bytes - this.bytes) * 8 / ((time - this.lastUpdateTime) * 1024);
	this.lastUpdateTime = time;
	this.updateSpeed(counter, bytes, speed);
	this.updateRate(rate);
	};

Queue.prototype.updateSpeed = function (counter, bytes, speed)
	{
	if (bytes != this.bytes || this.speeds[this.speeds.length - 1] != 0)
		{
		this.bytes = bytes;
		this.$speedInfo.text(speed.toFixed(1) + ' kbit/s');
		}
	this.addMissingSpeedRecords(counter);
	this.speeds.shift();
	this.speeds.push(speed);
	};

Queue.prototype.updateRate = function (rate)
	{
	if (rate != this.rate)
		{
		this.rate = rate;
		this.$rateInfo.text(rate >= maxSpeed - 1 ? 'unlimited' : rate + ' kbit/s');
		}
	};

Queue.prototype.addMissingSpeedRecords = function (counter)
	{
	if (counter <= this.lastCounter)
		return;
	while (++this.lastCounter < counter)
		{
		this.speeds.shift();
		this.speeds.push(0);
		}
	};

Queue.prototype.changeHistoryLength = function (length)
	{
	if (length < 2)
		return;
	this.speeds = reallocateArray(0, length, this.speeds);
	};

/*
 * User object
 */
function User(id, name,inIface,outIface,ips,crc)
	{
	this.id = id;
	this.name = name;
	this.ips = ips;
	this.crc = crc;
	this.inIface = inIface;
	this.outIface = outIface;
	this.uploadQueue = new Queue();
	this.downloadQueue = new Queue();

	this.updateCallback = null;
	}

User.prototype.createRow = function ()
	{
	var self = this;
	var $name_th = $('<th><a href="javascript:void(0)" class="name">' + this.name + '</a> <a href="javascript:void(0)" class="show-graph"><span>graf</span></a></th>');
	$name_th.find('a.name').click(function(ev) { showLimits(ev, self.id); });
	$name_th.find('a.show-graph').click(function() { createGraph(self.id); });
	var $ips = $('<td class="ips"><div title="' + this.ips + '">' + this.ips + '</div></div></td>');
	var $up_rate = $('<td class="up_rate">unlimited</td>');
	var $up_speed = $('<td class="up_speed">0.0 kbit/s</td>');
	var $down_rate = $('<td class="down_rate">unlimited</td>');
	var $down_speed = $('<td class="down_speed">0.0 kbit/s</td>');
	var $row = $('<tr class="item"></tr>');
	if (!this.inIface)
		{
		$down_rate.addClass('disabled');
		$down_speed.addClass('disabled');
		}
	else
		{
		$row.addClass(this.inIface);
		}
	if (!this.outIface)
		{
		$up_rate.addClass('disabled');
		$up_speed.addClass('disabled');
		}
	else
		{
		$row.addClass(this.outIface);
		}

	$row
		.append($name_th)
		.append($ips)
		.append($up_rate)
		.append($down_rate)
		.append($up_speed)
		.append($down_speed);

	this.$row = $row;
	this.uploadQueue.assignInfoElements($up_speed, $up_rate);
	this.downloadQueue.assignInfoElements($down_speed, $down_rate);

	return $row;
	};

User.prototype.onUpdate = function (callback)
	{
	this.updateCallback = callback;
	};

User.prototype.update = function (iface, counter, bytes, rate, time)
	{
	if (iface == this.inIface)
		this.downloadQueue.update(counter, bytes, rate, time);
	else if (iface == this.outIface)
		this.uploadQueue.update(counter, bytes, rate, time);
	else
		{
		console.log('Unknown iface: ' + iface);
		return;
		}
	if (this.updateCallback && this.uploadQueue.counter == this.downloadQueue.counter)
		this.updateCallback(this);
	};

User.prototype.getInterfaces = function ()
	{
	var ifaces = [];
	if (this.inIface)
		ifaces.push(this.inIface);
	if (this.outIface)
		ifaces.push(this.outIface);
	return ifaces;
	};

User.prototype.changeHistoryLength = function (length)
	{
	this.downloadQueue.changeHistoryLength(length);
	this.uploadQueue.changeHistoryLength(length);
	};

function reallocateArray(value, count, oldArray)
	{
	var array = new Array();
	if (oldArray)
		{
		while (count-- > 0 && oldArray.length > 0)
			array.unshift(oldArray.pop());
		}
	while (count-- > 0)
		array.unshift(value);
	return array;
	}

function updateStatistics()
	{
	for (var device in activeDevices)
		{
		$.get(statsUrl, { device: device, counter: nextCounter }, onXmlLoaded);
		}
	nextCounter++;
	}

function onXmlLoaded(data)
	{
	var $queues = $(data).find('queues');
	if ($queues.length == 0) return;
	var time = $queues.find('time').text();
	var device = $queues.find('device').text();
	var counter = parseInt($queues.find('counter').text());
	var $classes = $queues.find('class');

	$classes.each(function ()
		{
		var $Class = $(this);
		var id = $Class.attr('id');
		if (id != '0' && id != 'ffff' && id != 'fffe')
			{
			var rate = parseInt($Class.find('rate').text());
			var bytes = parseInt($Class.find('bytes').text());

			users[id].update(device, counter, bytes, rate, time);
			}
		});
	}

var updateTimer = 0;
var activeDevices = {};
var linkedDevices = {};

function initSettingsPanel()
	{
	$('#changeIntervalButton').click(setTimer);
	$('#pauseButton').click(pause);
	$('#addQueueButton').click(function (event) { showLimits(event, -1); });
	$('#interval').spinner({ min: 1 });
	}

function setTimer()
	{
	updateStatistics();
	if (updateTimer) clearInterval(updateTimer);
	updateTimer = window.setInterval("updateStatistics()", $('#interval').val() * 1000);
	}

function pause()
	{
	var $button = $('#pauseButton');
	if (updateTimer)
		{
		clearInterval(updateTimer);
		updateTimer = null;
		$button.val('Spustit');
		}
	else
		{
		setTimer();
		$button.val('Pozastavit');
		}
	}

function fillQueuesList()
	{
	var $queues_list = $('#queues-list tbody');
	$.each(users, function(idx, user)
		{
		$queues_list.append(user.createRow());
		});
	}

function initFilters()
	{
	var template = '<label><input id="#" type="checkbox" checked="checked" /><span>#</span></label>';
	var $filters = $('#filters');
	for (var idx = 0; idx < devices.length; ++idx)
		{
		activeDevices[devices[idx]] = true;
		linkedDevices[devices[idx]] = {};
		$filters.append(template.replace(/#/g, devices[idx]));
		}
	$filters.find(':checkbox').change(function ()
		{
		$('.filtered').toggleClass($(this).attr('id'));
		refreshActiveDevices();
		});
	for (var id in users)
		{
		if (!users[id].inIface || !users[id].outIface)
			continue;
		linkedDevices[users[id].inIface][users[id].outIface] = true;
		linkedDevices[users[id].outIface][users[id].inIface] = true;
		}
	$('.filtered').addClass(devices.join(' '));
	}

function refreshActiveDevices()
	{
	var newActiveDevices = {};
	var $checked = $('#filters :checked').each(function ()
		{
		var device = $(this).attr('id');
		newActiveDevices[device] = true;
		$.extend(newActiveDevices, linkedDevices[device]);
		});
	activeDevices = newActiveDevices;
	}

function initLimitsDialog()
	{
	var $in_iface = $('#in_iface'),
		$out_iface = $('#out_iface');
	for (var idx = 0; idx < devices.length; ++idx)
		{
		var option = '<option value="' + devices[idx] + '">' + devices[idx] + '</option>';
		$in_iface.append(option);
		$out_iface.append(option);
		}
	$('#limits-dialog .close').click(closeLimits);
	$('#limits-dialog #delete').click(function ()
		{
		$('#delete-form').submit();
		});
	}

function showLimits(e, id)
	{
	var posx = 0;
	var posy = 0;
	if (!e)
		{
		e = window.event;
		}
	if (e.clientX || e.clientY)
		{
		posx = e.clientX;
		posy = e.clientY;
		}

	var $limits = $('#limits-dialog');
	$limits
		.css('top', (posy + 8) + 'px')
		.show();
	if (id > 0)
		{
		$('#title').text(users[id].name);
		$('#name, #old_name, #delete_old_name').val(users[id].name);
		$('#id, #delete_id').val(id);
		$('#upload').val(users[id].uploadQueue.rate);
		$('#download').val(users[id].downloadQueue.rate);
		$('#in_iface').val(users[id].inIface);
		$('#out_iface').val(users[id].outIface);
		$('#ips').val(users[id].ips);
		$('#crc').val(users[id].crc);
		$('#delete').removeAttr('disabled');
		}
	else
		{
		$('#title').text('Nová fronta');
		$('#name').val('Jméno');
		$('#old_name').val('');
		$('#id').val(-1);
		$('#upload, #download, #crc').val(0);
		$('#in_iface, #out_iface, #ips').val('');
		$('#delete').attr('disabled', 'disabled');
		}
	}

function closeLimits()
	{
	$('#limits-dialog').hide();
	}

function createGraph(userId)
	{
	var userGraphClass = 'user' + userId;
	if ($('#graphs .graph.' + userGraphClass).length)
		return;
	var user = users[userId];
	var $graph = $('<div class="graph item ' + userGraphClass + ' ' + user.getInterfaces().join(' ') + '"></div>');
	var graphObj = new Graph($graph[0], user);
	$graph.prepend('<h4>' + user.name + '<a href="javascript:void(0)" class="close cross">x</a></h4>');
	$('#graphs').append($graph);
	graphObj.register();
	$graph.find('a.close').click(function ()
		{
		graphObj.unregister();
		$graph.remove();
		});
	$graph.data('graph', graphObj);
	}

/*
 * Graph object
 */
function Graph(parent, user)
	{
	this.user = user;
	this.minUnit = 16; // kB/s
	this.width = 400;
	this.height = 200;
	this.canvas = Raphael(parent, this.width + 1, this.height + 1);
	this.leftMargin = 40;
	this.topMargin = 5;
	this.yAxisMaximum = 0;

	this.draw();
	}

Graph.prototype.register = function ()
	{
	var self = this;
	this.user.onUpdate(function () { self.draw(); });
	};

Graph.prototype.unregister = function ()
	{
	this.user.onUpdate(null);
	};

Graph.prototype.draw = function ()
	{
	var newMaxYAxis = this.getYAxisMaximum();
	if (this.yAxisMaximum < newMaxYAxis || this.yAxisMaximum / 2 > newMaxYAxis)
		{
		this.canvas.clear();
		this.yAxisMaximum = newMaxYAxis;
		this.drawGrid();
		}
	else if (this.downloadPoints)
		{
		this.downloadPoints.remove();
		this.uploadPoints.remove();
		}
	this.downloadPoints = this.drawPoints(this.user.downloadQueue.speeds, '#5d5');
	this.uploadPoints = this.drawPoints(this.user.uploadQueue.speeds, '#d55');
	};

Graph.prototype.getYAxisMaximum = function ()
	{
	var maxDownload = Math.max.apply(Math, this.user.downloadQueue.speeds);
	var maxUpload = Math.max.apply(Math, this.user.uploadQueue.speeds);
	return Math.ceil(Math.max.apply(Math, [ maxDownload * 1.2, maxUpload * 1.2, this.minUnit ]) / this.minUnit) * this.minUnit;
	};

Graph.prototype.drawGrid = function ()
	{
	var path = [ 'M', this.leftMargin + 0.5, 0.5, 'V', this.height + 0.5, 'H', this.width + 0.5 ];
	this.canvas.path(path.join()).attr({ stroke: '#aaa' });
	var labelDensity = this.computeLabelDensity();
	path = [];
	for (var offset = labelDensity; offset < this.yAxisMaximum; offset += labelDensity)
		{
		path = path.concat([ 'M', this.leftMargin + 0.5, this.getY(offset), 'H', this.width + 0.5 ]);
		this.drawLabel(offset);
		}
	this.canvas.path(path.join()).attr({ stroke: '#ddd' });
	};

Graph.prototype.computeLabelDensity = function ()
	{
	var max = this.yAxisMaximum;
	var density = this.minUnit / 4;
	while (max / density > 8)
		{
		density *= 2;
		}
	return density;
	};

Graph.prototype.drawLabel = function (speed)
	{
	this.canvas
		.text(this.leftMargin / 2, this.getY(speed), Math.round(speed) + 'k')
		.attr({ font: '10px Helvetica, Arial', fill: "#aaa" });
	};

Graph.prototype.drawPoints = function (points, color)
	{
	var count = points.length;
	var set = this.canvas.set();
	var path = [ 'M', this.leftMargin + 0.5, this.getY(points[0]) ];
	var width = this.width - this.leftMargin;
	for (var idx = 1; idx < count; ++idx)
		{
		var x = Math.round(idx * width / (count - 1)) + this.leftMargin + 0.5,
			y = this.getY(points[idx]);
		path = path.concat([ 'L', x, y ]);
		set.push(this.canvas.rect(x - 1.5, y - 1.5, 3, 3).attr({ fill: color }));
		}
	set.push(this.canvas.path(path.join()));
	return set.attr({ stroke: color });
	};

Graph.prototype.getY = function (y)
	{
	var height = this.height - this.topMargin;
	return Math.round(height - y * (height / this.yAxisMaximum)) + this.topMargin + 0.5;
	};
